from __future__ import annotations

import base64
import json
import os
from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional

from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.sql._internal.sql_source import BaseSQLSource, SQLSourceKind
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from sqlalchemy.engine import Engine
    from sqlalchemy.engine.url import URL


_logger = get_logger(__name__)


_SPANNER_PROJECT_NAME = "SPANNER_PROJECT"
_SPANNER_INSTANCE_NAME = "SPANNER_INSTANCE"
_SPANNER_DATABASE_NAME = "SPANNER_DATABASE"
_SPANNER_CREDENTIALS_BASE64_NAME = "SPANNER_CREDENTIALS_BASE64"
_SPANNER_EMULATOR_HOST_NAME = "SPANNER_EMULATOR_HOST"
_SPANNER_ENGINE_ARGUMENTS_NAME = "SPANNER_ENGINE_ARGUMENTS"


class SpannerSourceImpl(BaseSQLSource):
    kind = SQLSourceKind.spanner

    def __init__(
        self,
        name: str | None = None,
        project: str | None = None,
        instance: str | None = None,
        database: str | None = None,
        credentials_base64: str | None = None,
        emulator_host: str | None = None,
        engine_args: Dict[str, Any] | None = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        try:
            from google.cloud import sqlalchemy_spanner
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[spanner]")
        del sqlalchemy_spanner  # unused

        self.ingested_tables: Dict[str, Any] = {}
        self.project_id = project or load_integration_variable(
            integration_name=name, name=_SPANNER_PROJECT_NAME, override=integration_variable_override
        )
        self.instance_id = instance or load_integration_variable(
            integration_name=name, name=_SPANNER_INSTANCE_NAME, override=integration_variable_override
        )
        self.database_id = database or load_integration_variable(
            integration_name=name, name=_SPANNER_DATABASE_NAME, override=integration_variable_override
        )
        self.credentials_base64 = credentials_base64 or load_integration_variable(
            integration_name=name, name=_SPANNER_CREDENTIALS_BASE64_NAME, override=integration_variable_override
        )
        self.emulator_host = emulator_host or load_integration_variable(
            integration_name=name, name=_SPANNER_EMULATOR_HOST_NAME, override=integration_variable_override
        )

        eargs = {}
        if engine_args is not None:
            eargs = engine_args
        else:
            found = load_integration_variable(
                integration_name=name,
                name=_SPANNER_ENGINE_ARGUMENTS_NAME,
                parser=json.loads,
                override=integration_variable_override,
            )
            if found is not None:
                eargs = found

        # The SQL sources are read-only, so transactions are not needed.
        # Setting the isolation level on the engine, instead of the connection, avoids
        # a DBAPI statement to reset the transactional level back to the default before returning the
        # connection to the pool.
        eargs.setdefault(
            "isolation_level",
            os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"),
        )

        # Spanner defaults to ReadWrite, but Chalk only reads data from spanner.
        eargs.setdefault("read_only", True)

        BaseSQLSource.__init__(self, name=name, engine_args=eargs, async_engine_args={})

    # Spanner does not support using URL, so we need to create the engine directly (rather than using URL)
    def get_engine(self) -> "Engine":
        try:
            from google.auth.credentials import AnonymousCredentials
            from google.cloud import spanner
            from google.oauth2 import service_account
            from sqlalchemy.engine import create_engine
        except ImportError:
            raise missing_dependency_exception("chalkpy[spanner]")

        if self._engine is None:  # pyright: ignore[reportUnnecessaryComparison]
            connect_args = {}

            # Uses workload identity if credentials are not provided
            if self.credentials_base64:
                decoded_credentials = base64.b64decode(self.credentials_base64).decode("utf-8")
                json_acct_info: Dict = json.loads(decoded_credentials)
                credentials = (
                    service_account.Credentials.from_service_account_info(json_acct_info)
                    if not self.emulator_host
                    else AnonymousCredentials()
                )
                connect_args = dict(
                    client=spanner.Client(
                        project=self.project_id,
                        client_options={"api_endpoint": self.emulator_host} if self.emulator_host else None,
                        credentials=credentials,
                    ),
                )
            elif self.emulator_host:
                connect_args = dict(
                    client=spanner.Client(
                        project=self.project_id,
                        client_options={"api_endpoint": self.emulator_host},
                        credentials=AnonymousCredentials(),
                    )
                )

            self._engine = create_engine(
                self.local_engine_url().render_as_string(),  # Spanner adapter does not support URL
                connect_args=connect_args,
            ).execution_options(**self.engine_args)

        return self._engine

    def local_engine_url(self) -> "URL":
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="spanner+spanner",
            host=f"/projects/{self.project_id}/instances/{self.instance_id}",
            database=f"databases/{self.database_id}",
        )

    def get_sqlglot_dialect(self) -> str | None:
        # BigQuery and Spanner have very similar dialects (but not identical)
        return "bigquery"

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(_SPANNER_PROJECT_NAME, self.name, self.project_id),
                create_integration_variable(_SPANNER_INSTANCE_NAME, self.name, self.instance_id),
                create_integration_variable(_SPANNER_DATABASE_NAME, self.name, self.database_id),
                create_integration_variable(_SPANNER_CREDENTIALS_BASE64_NAME, self.name, self.credentials_base64),
                create_integration_variable(_SPANNER_EMULATOR_HOST_NAME, self.name, self.emulator_host),
            ]
            if v is not None
        }



from __future__ import annotations

from chalk._gen.chalk.auth.v1.permissions_pb2 import Permission
from chalk.client.client import ChalkAPIClientProtocol, ChalkClient
from chalk.client.client_async import AsyncChalkClient
from chalk.client.exc import ChalkAuthException, ChalkBaseException
from chalk.client.models import (
    BranchDeployResponse,
    BranchIdParam,
    BulkOnlineQueryResponse,
    ChalkError,
    ChalkException,
    DatasetFilter,
    ErrorCode,
    ErrorCodeCategory,
    FeatureDropResponse,
    FeatureObservationDeletionResponse,
    FeatureReference,
    FeatureResult,
    GetIncrementalProgressResponse,
    OfflineQueryContext,
    OnlineQuery,
    OnlineQueryContext,
    OnlineQueryResponse,
    PlanQueryResponse,
    QueryMeta,
    QueryStatus,
    ResolverRunResponse,
    ResourceRequests,
    StreamResolverTestResponse,
    WhoAmIResponse,
)
from chalk.client.response import Dataset, DatasetRevision, OnlineQueryResult

__all__ = [
    "AsyncChalkClient",
    "BranchDeployResponse",
    "BranchIdParam",
    "BulkOnlineQueryResponse",
    "ChalkAPIClientProtocol",
    "ChalkAuthException",
    "ChalkBaseException",
    "ChalkClient",
    "ChalkError",
    "ChalkException",
    "Dataset",
    "DatasetFilter",
    "DatasetRevision",
    "ErrorCode",
    "ErrorCodeCategory",
    "FeatureDropResponse",
    "FeatureObservationDeletionResponse",
    "FeatureReference",
    "FeatureResult",
    "GetIncrementalProgressResponse",
    "OfflineQueryContext",
    "OnlineQuery",
    "OnlineQueryContext",
    "OnlineQueryResponse",
    "OnlineQueryResult",
    "Permission",
    "PlanQueryResponse",
    "QueryMeta",
    "QueryStatus",
    "ResolverRunResponse",
    "ResourceRequests",
    "StreamResolverTestResponse",
    "WhoAmIResponse",
]



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.server.v1 import topic_pb2 as _topic_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class ScheduledJob(_message.Message):
    __slots__ = ("id", "name", "payload", "schedule", "topic", "attributes", "environment", "tags")
    class AttributesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    SCHEDULE_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    payload: str
    schedule: str
    topic: _topic_pb2.Topic
    attributes: _containers.ScalarMap[str, str]
    environment: str
    tags: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self,
        id: _Optional[str] = ...,
        name: _Optional[str] = ...,
        payload: _Optional[str] = ...,
        schedule: _Optional[str] = ...,
        topic: _Optional[_Union[_topic_pb2.Topic, _Mapping]] = ...,
        attributes: _Optional[_Mapping[str, str]] = ...,
        environment: _Optional[str] = ...,
        tags: _Optional[_Iterable[str]] = ...,
    ) -> None: ...

class UpdateJobRequest(_message.Message):
    __slots__ = ("id", "name", "payload", "schedule", "topic", "attributes", "environment", "tags")
    class AttributesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    SCHEDULE_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    payload: str
    schedule: str
    topic: _topic_pb2.Topic
    attributes: _containers.ScalarMap[str, str]
    environment: str
    tags: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self,
        id: _Optional[str] = ...,
        name: _Optional[str] = ...,
        payload: _Optional[str] = ...,
        schedule: _Optional[str] = ...,
        topic: _Optional[_Union[_topic_pb2.Topic, _Mapping]] = ...,
        attributes: _Optional[_Mapping[str, str]] = ...,
        environment: _Optional[str] = ...,
        tags: _Optional[_Iterable[str]] = ...,
    ) -> None: ...

class UpdateJobResponse(_message.Message):
    __slots__ = ("job",)
    JOB_FIELD_NUMBER: _ClassVar[int]
    job: ScheduledJob
    def __init__(self, job: _Optional[_Union[ScheduledJob, _Mapping]] = ...) -> None: ...

class ListJobsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListJobsResponse(_message.Message):
    __slots__ = ("jobs",)
    JOBS_FIELD_NUMBER: _ClassVar[int]
    jobs: _containers.RepeatedCompositeFieldContainer[ScheduledJob]
    def __init__(self, jobs: _Optional[_Iterable[_Union[ScheduledJob, _Mapping]]] = ...) -> None: ...

class CreateJobRequest(_message.Message):
    __slots__ = ("name", "payload", "schedule", "topic", "attributes", "environment", "tags")
    class AttributesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    NAME_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    SCHEDULE_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    name: str
    payload: str
    schedule: str
    topic: _topic_pb2.Topic
    attributes: _containers.ScalarMap[str, str]
    environment: str
    tags: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self,
        name: _Optional[str] = ...,
        payload: _Optional[str] = ...,
        schedule: _Optional[str] = ...,
        topic: _Optional[_Union[_topic_pb2.Topic, _Mapping]] = ...,
        attributes: _Optional[_Mapping[str, str]] = ...,
        environment: _Optional[str] = ...,
        tags: _Optional[_Iterable[str]] = ...,
    ) -> None: ...

class CreateJobResponse(_message.Message):
    __slots__ = ("job",)
    JOB_FIELD_NUMBER: _ClassVar[int]
    job: ScheduledJob
    def __init__(self, job: _Optional[_Union[ScheduledJob, _Mapping]] = ...) -> None: ...

class DeleteJobRequest(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class DeleteJobResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetJobByNameRequest(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class GetJobByNameResponse(_message.Message):
    __slots__ = ("job",)
    JOB_FIELD_NUMBER: _ClassVar[int]
    job: ScheduledJob
    def __init__(self, job: _Optional[_Union[ScheduledJob, _Mapping]] = ...) -> None: ...



from __future__ import annotations

import inspect
import traceback
from typing import TYPE_CHECKING, Mapping, Sequence

from chalk.features import unwrap_feature
from chalk.utils.object_inspect import get_source_object_starting
from chalk.utils.source_parsing import should_skip_source_code_parsing

if TYPE_CHECKING:
    from chalk.client.models import FeatureReference

MAX_NAMED_QUERY_FUNCTION_DEF_LEN = 400


class NamedQuery:
    def __init__(
        self,
        *,
        name: str,
        version: str | None = None,
        input: Sequence[FeatureReference] | None = None,
        output: Sequence[FeatureReference] | None = None,
        tags: Sequence[str] | None = None,
        description: str | None = None,
        owner: str | None = None,
        meta: Mapping[str, str] | None = None,
        staleness: Mapping[FeatureReference, str] | None = None,
        planner_options: Mapping[str, str | int | bool] | None = None,
        additional_logged_features: Sequence[FeatureReference] | None = None,
        valid_plan_not_required: bool = False,
    ):
        """Create a named query.

        Named queries are aliases for specific queries that can be used by API clients.

        Parameters
        ----------
        name
            A name for the named query—this can be versioned with the version parameter, but
            must otherwise be unique. The name of the named query shows up in the dashboard and
            is used to specify the outputs for a query.
        version:
            A string specifying the version of the named query: version is not required, but
            if specified it must be a valid "semantic version".
        input
            The features which will be provided by callers of this query.
            For example, `[User.id]`. Features can also be expressed as snakecased strings,
            e.g. `["user.id"]`.
        output
            Outputs are the features that you'd like to compute from the inputs.
            For example, `[User.age, User.name, User.email]`.

            If an empty sequence, the output will be set to all features on the namespace
            of the query. For example, if you pass as input `{"user.id": 1234}`, then the query
            is defined on the `User` namespace, and all features on the `User` namespace
            (excluding has-one and has-many relationships) will be used as outputs.
        tags
            Allows selecting resolvers with these tags.
        description
            A description of the query. Rendered in the Chalk UI and used for search indexing.
        owner
            The owner of the query. This should be a Slack username or email address.
            This is used to notify the owner in case of incidents
        meta
            Additional metadata for the query.
        staleness
            Maximum staleness overrides for any output features or intermediate features.
            See https://docs.chalk.ai/docs/query-caching for more information.
        planner_options
            Dictionary of additional options to pass to the Chalk query engine.
            Values may be provided as part of conversations with Chalk support
            to enable or disable specific functionality.
        additional_logged_features
            If query logging is enabled, this will log the specified features in addition to
            the output features to the query log.
        valid_plan_not_required
            If False (which is the default), new main deployments will fail if the planner cannot
            generate a valid plan for the named query, preventing rollover of a bad deployment. If
            set to True, the query will not be preplanned and therefore will not be validated on
            deployment.


        Examples
        --------
        >>> from chalk import NamedQuery
        >>> # this query's name and version can be used to specify query outputs in an API request.
        >>> NamedQuery(
        ...     name="fraud_model",
        ...     version="1.0.0",
        ...     input=[User.id],
        ...     output=[User.age, User.fraud_score, User.credit_report.fico],
        ... )
        """
        super().__init__()
        self.errors = []

        filename = None
        source_line_start = None
        source_line_end = None
        source_code = None

        if not should_skip_source_code_parsing():
            try:
                internal_frame = inspect.currentframe()
                if internal_frame is not None:
                    definition_frame = internal_frame.f_back
                    if definition_frame is not None:
                        filename = definition_frame.f_code.co_filename
                        source_line_start = definition_frame.f_lineno
                        source_code, source_line_start, source_line_end = get_source_object_starting(definition_frame)
                    del internal_frame
            except Exception:
                pass

        self._input = None
        self._output = None
        self._additional_logged_features = None
        self._input_raw = input
        self._output_raw = output
        self._additional_logged_features_raw = additional_logged_features
        self.name = name
        self.version = version and str(version)
        self.tags = [str(t) for t in tags] if tags else None
        self.filename = filename
        self.description = description
        self.owner = owner
        self.meta = meta
        self.staleness = {str(k): v for k, v in staleness.items()} if staleness else None
        self.planner_options = {k: str(v) for k, v in planner_options.items()} if planner_options else None
        self.source_line_start = source_line_start
        self.code = source_code
        self.source_line_end = source_line_end
        self.valid_plan_not_required = valid_plan_not_required

        dup_nq = NAMED_QUERY_REGISTRY.get((name, version), None)
        if dup_nq is not None:
            self.errors.append(
                (
                    "Named query must be distinct on name and version, but found two named queries with name "
                    f"'{name}' and version '{version} in files '{dup_nq.filename}' and '{filename}'."
                )
            )

        NAMED_QUERY_REGISTRY[(name, version)] = self

    @property
    def input(self):
        if self._input is not None:
            return self._input
        try:
            if self._input_raw is None and self._output_raw is None:
                self._input = None
                self.errors.append("Must provide either input or output ")
            elif self._input_raw is not None:
                self._input = [str(f) for f in self._input_raw]
            elif self._output_raw is not None:
                self._input = [str(unwrap_feature(o).primary_feature) for o in self._output_raw]
        except Exception:
            self._input = None
            self.errors.append(traceback.format_exc())

        return self._input

    @property
    def output(self):
        if self._output is not None:
            return self._output
        try:
            if self._output_raw is not None:
                self._output = [str(o) for o in self._output_raw]
        except Exception:
            self._output = None
            self.errors.append(f"Error creating NamedQuery '{self.name} ({self.version})': {traceback.format_exc()}")

        return self._output

    @property
    def additional_logged_features(self):
        if self._additional_logged_features is not None:
            return self._additional_logged_features
        try:
            if self._additional_logged_features_raw is not None:
                self._additional_logged_features = [str(alf) for alf in self._additional_logged_features_raw]
        except Exception:
            self._additional_logged_features = None
            self.errors.append(f"Error creating NamedQuery '{self.name} ({self.version})': {traceback.format_exc()}")

        return self._additional_logged_features


NAMED_QUERY_REGISTRY: dict[tuple[str, str | None], NamedQuery] = {}



import collections.abc
from dataclasses import is_dataclass
from typing import TYPE_CHECKING, Any, FrozenSet, List, Mapping, Union

import attrs
import pandas as pd
import pyarrow as pa

from chalk import DataFrame
from chalk.client import ChalkBaseException, ChalkError, ErrorCode
from chalk.features import Feature
from chalk.utils.missing_dependency import missing_dependency_exception
from chalk.utils.pydanticutil.pydantic_compat import is_pydantic_basemodel_instance

if TYPE_CHECKING:
    import polars as pl


class _UploadFeaturesValidationErrors:
    @staticmethod
    def expected_mapping(e: Any):
        return ChalkError.create(
            code=ErrorCode.INVALID_QUERY,
            message=(
                f"Multi upload detected a list of items, but some list items are `{type(e).__name__}`. If you "
                "wish to upload a list of items, be sure each item in the list is a mapping where "
                "each key is a feature and each value is the desired value for that feature"
                "in context of the other inputs specified by the other key:value pairs in the "
                "mapping."
            ),
        )

    @staticmethod
    def expected_same_length_lists(
        expected_length: int, expected_length_col: str, actual_length: int, actual_length_col: str
    ):
        expected_s = "" if expected_length == 1 else "s"
        actual_s = "" if actual_length == 1 else "s"
        return ChalkError.create(
            code=ErrorCode.INVALID_QUERY,
            message=(
                f"expected all values in mapping to be lists of the same length, found {expected_length} "
                f"value{expected_s} for '{expected_length_col}' and {actual_length} value{actual_s} for {actual_length_col}"
            ),
        )

    @staticmethod
    def expected_list_in_mapping(col_name: str, col_value: Any):
        return ChalkError.create(
            code=ErrorCode.INVALID_QUERY,
            message=f"expected all values in mapping to be lists, found '{type(col_value).__name__}' for '{col_name}'",
        )

    @staticmethod
    def structs_not_supported(fqn: str, obj: Any) -> ChalkError:
        return ChalkError.create(
            code=ErrorCode.INVALID_QUERY,
            message=(
                f"Multi-upload does not yet support uploading structs, "
                f"but values for feature '{fqn}' has the type {type(obj).__name__}"
            ),
        )


def _is_struct_like_object(obj: Any) -> bool:
    return (
        is_dataclass(obj)
        or (isinstance(obj, tuple) and hasattr(obj, "_fields"))
        or is_pydantic_basemodel_instance(obj)
        or attrs.has(obj.__class__)
        or isinstance(obj, Mapping)
    )


def _validate_mapping(inputs: Mapping[Union[str, Feature, Any], Any]) -> List[ChalkError]:
    if len(inputs) == 0:
        return []

    num_rows = None
    reference_col = None
    for k, v in inputs.items():
        if not isinstance(v, list):
            return [_UploadFeaturesValidationErrors.expected_list_in_mapping(str(k), v)]
        if num_rows is None:
            reference_col = k
            num_rows = len(v)
        elif num_rows != len(v):
            assert reference_col is not None
            return [
                _UploadFeaturesValidationErrors.expected_same_length_lists(num_rows, str(reference_col), len(v), str(k))
            ]
        if _is_struct_like_object(v[0]):
            return [_UploadFeaturesValidationErrors.structs_not_supported(str(k), v[0])]

    return []


def _validate_list(inputs: List) -> List[ChalkError]:
    if len(inputs) == 0:
        return []

    for e in inputs:
        if not isinstance(e, collections.abc.Mapping):
            return [_UploadFeaturesValidationErrors.expected_mapping(e)]

        for k, v in e.items():
            if _is_struct_like_object(v):
                return [_UploadFeaturesValidationErrors.structs_not_supported(k, v)]

    return []


def to_multi_upload_inputs(
    inputs: Union[
        List[Mapping[Union[str, Feature, Any], Any]],
        Mapping[Union[str, Feature, Any], List[Any]],
        pd.DataFrame,
        "pl.DataFrame",
        DataFrame,
    ],
) -> List[pa.Table]:
    tables: List[pa.Table] = []
    if isinstance(inputs, DataFrame):
        tables.append(inputs.to_pyarrow())
    elif isinstance(inputs, collections.abc.Mapping):
        if len(inputs) == 0:
            return []
        errs = _validate_mapping(inputs)
        if errs:
            raise ChalkBaseException(errs, detail="multi-upload input validation failed")
        tables.append(pa.Table.from_pydict({str(k): lv for (k, lv) in inputs.items()}))
    elif isinstance(inputs, list):
        if len(inputs) == 0:
            return []
        errs = _validate_list(inputs)
        if errs:
            raise ChalkBaseException(errs, detail="multi-upload input validation failed")
        common_keys_to_mappings: Mapping[FrozenSet, List[Mapping]] = collections.defaultdict(list)
        for mapping in inputs:
            common_keys_to_mappings[frozenset(mapping.keys())].append(mapping)
        for common_mappings_list in common_keys_to_mappings.values():
            fqn_to_values = collections.defaultdict(list)
            for mapping in common_mappings_list:
                for k, v in mapping.items():
                    fqn_to_values[k].append(v)
            tables.append(pa.Table.from_pydict(fqn_to_values))
    elif isinstance(inputs, pd.DataFrame):
        tables.append(pa.Table.from_pandas(inputs))
    else:
        try:
            import polars as pl
        except ImportError:
            raise missing_dependency_exception("chalkpy[runtime]")
        if isinstance(inputs, (pl.DataFrame, pl.LazyFrame)):  # pyright: ignore[reportUnnecessaryIsInstance]
            tables.append(inputs.to_arrow())
        else:
            try:
                df = pd.DataFrame(inputs)
            except:
                raise ValueError(
                    f"Multi upload received an input type that couldn't be interpreted: {str(type(inputs))}."
                )
            tables.append(pa.Table.from_pandas(df))

    if not tables:
        raise ValueError("Multi upload received an empty `inputs` object")

    return tables



from __future__ import annotations

import contextvars
import inspect
import textwrap
from collections import defaultdict
from datetime import timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    cast,
    Set,
    Protocol,
    Mapping,
)

from typing_extensions import TypeGuard

from chalk._lsp._class_finder import get_class_ast
from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
from chalk.utils import notebook
from chalk.utils.collections import ensure_tuple, get_unique_item
from chalk.utils.source_parsing import should_skip_source_code_parsing

if TYPE_CHECKING:
    from chalk.features.feature_field import Feature
    from chalk._lsp.error_builder import FeatureClassErrorBuilder
    from chalk.features.underscore_features import NamedUnderscoreExpr

__all__ = ["Features", "FeaturesMeta", "FeatureSetBase", "is_features_cls", "is_feature_set_class"]


class ClassSource:
    __slots__ = ("filename", "class_source", "tree", "dedent_source")

    def __init__(self, c: Any):
        self.filename = None
        self.class_source = None
        self.tree = None
        super().__init__()
        if not should_skip_source_code_parsing():
            try:
                self.filename = inspect.getfile(c)
            except:
                self.filename = None

            try:
                if notebook.is_defined_in_module(c):
                    self.class_source = inspect.getsource(c)
                # TODO (rkargon) Try to get source from notebook cells if possible
                else:
                    self.class_source = None
            except:
                self.class_source = None
            self.dedent_source = self.class_source and textwrap.dedent(self.class_source)
            try:
                self.tree = get_class_ast(c)
            except:
                self.tree = None


class FeaturesMeta(type):
    """Metaclass for classes decorated with ``@features``.

    This metaclass allows for:

    1.  Classes annotated with @features to pass the
        ``isinstance(x, Features)`` and ``issubclass(X, Features)`` checks.
    2.  ``Features[Feature1, Feature2]`` annotations to return subclasses of Features, so they can be used as proper type annotations.
    """

    _features_lazy: Callable[[], List[Feature]]  # pyright: ignore[reportUninitializedInstanceVariable]

    @property
    def features(self) -> List[Feature]:
        return self._features_lazy()

    def __subclasscheck__(self, subcls: type) -> bool:
        if not isinstance(subcls, type):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"{subcls} is not a type")
        if hasattr(subcls, "__is_features__"):
            return getattr(subcls, "__is_features__")
        return False

    def __instancecheck__(self, instance: object) -> bool:
        return self.__subclasscheck__(type(instance))

    def __getitem__(cls, item: Any) -> Any:
        # This lets us subscript by the features
        # Annotating the return type as Type[Any] as instances of @features classes are
        # not recognized as being subclasses of Features by the type checker (even though at runtime they would be)
        from chalk.features.dataframe import DataFrame
        from chalk.features.feature_field import Feature

        is_exploded = False

        # Typing the `__getitem__` as any, since the @features members are typed as the underlying data structure
        # But, they should always be features or a tuple of features
        if isinstance(item, type) and issubclass(item, Features):
            is_exploded = True
            item = [
                f
                for f in item.features
                if not f.is_autogenerated and not f.is_has_one and not f.is_has_many and not f.is_windowed
            ]

        def f():
            item_ = ensure_tuple(item)
            item_ = tuple(unwrap_feature(x) if isinstance(x, FeatureWrapper) else x for x in item_)
            for x in item_:
                if isinstance(x, str):
                    raise TypeError(
                        (
                            f'String features like {cls.__name__}["{x}"] are unsupported. '
                            f'Instead, replace with "{cls.__name__}[{x}]"'
                        )
                    )

                if isinstance(x, Feature) or (isinstance(x, type) and issubclass(x, DataFrame)):  # pyright: ignore[reportUnnecessaryIsInstance]
                    continue

                raise TypeError(f"Invalid feature {x} of type '{type(x).__name__}'")
            cls_ = cast(Type[Features], cls)

            new_features = list([*getattr(cls_, "features", ()), *item_])
            if len(new_features) == 0:
                raise ValueError("Cannot create empty Features[] class")
            return new_features

        class SubFeatures(cls):  # pyright: ignore
            __is_exploded__ = is_exploded
            _features_lazy = f

        return SubFeatures

    def __repr__(cls) -> str:
        if not hasattr(cls, "features"):
            return "Features"
        return f"Features[{', '.join(str(f) for f in cls.features)}]"

    def __call__(cls, *args: object, **kwargs: object):
        raise RuntimeError("Instances features cls should never be directly created. Instead, use Features[User.id]")

    def __sub__(self, other: type[Features] | FeatureWrapper):
        if isinstance(other, type) and issubclass(other, Features):  # pyright: ignore[reportUnnecessaryIsInstance]
            other_features = set(other.features)
            return FeaturesImpl[tuple(f for f in self.features if f not in other_features)]
        if isinstance(other, FeatureWrapper):  # pyright: ignore[reportUnnecessaryIsInstance]
            return FeaturesImpl[tuple(f for f in self.features if f != unwrap_feature(other))]
        raise TypeError(f"Invalid type {type(other)} for subtraction with Features[...]")

    @property
    def namespace(cls):
        cls = cast(Type[Features], cls)
        return get_unique_item((x.namespace for x in cls.features), name=f"{cls.__name__} feature namespaces")


class FeaturesImpl(metaclass=FeaturesMeta):
    """Features base class.

    This class is never instantiated or directly inherited. However, classes
    annotated with @features can be thought of as inheriting from this class.
    It can be used with `isinstance` and `issubclass` checks, as well as for
    typing.
    """

    # Internally, the Features class is instantiated when results come through, and
    # results are bound to instances of this class via attributes
    __chalk_error_builder__: ClassVar[FeatureClassErrorBuilder]
    __chalk_is_singleton__: ClassVar[bool]
    __chalk_etl_offline_to_online__: ClassVar[bool]
    __chalk_max_staleness__: ClassVar[timedelta]
    __chalk_namespace__: ClassVar[str]
    __chalk_primary__: ClassVar[Feature | None] = None  # The primary key feature
    __chalk_owner__: ClassVar[Optional[str]]
    __chalk_tags__: ClassVar[List[str]]
    __chalk_ts__: ClassVar[Optional[Feature]]  # The timestamp feature
    features: ClassVar[List[Feature]]
    __chalk_features_raw__: ClassVar[List[Feature]]
    __chalk_source_info__: "ClassVar[ClassSource]"
    __chalk_is_loaded_from_notebook__: ClassVar[bool] = False
    """True if loaded on a branch server from a pickle"""
    __chalk_materialized_windows__: ClassVar[List[Feature]]
    __chalk_group_by_materialized_windows__: ClassVar[List[Feature]]
    __chalk_expression_windows__: ClassVar[List[Feature]]
    __chalk_notebook_feature_expressions__: ClassVar[Dict[str, NamedUnderscoreExpr]]

    __is_exploded__: ClassVar[bool] = False
    __is_features__: ClassVar[bool] = True

    # When constructing results, this class is instantiated

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        """Iterating over features yields tuples of (fqn, value) for all scalarish feature values."""
        raise NotImplementedError

    def __len__(self) -> int:
        """The number of features that are set."""
        raise NotImplementedError

    def items(self) -> Iterator[Tuple[str, Any]]:
        """Iterating over features yields tuples of (fqn, value) for all feature values."""
        raise NotImplementedError


if TYPE_CHECKING:

    class FeaturesProtocol(FeaturesImpl):
        def __class_getitem__(cls, item: Any) -> "FeaturesProtocol": ...

    Features = FeaturesProtocol
else:
    Features = FeaturesImpl


class FeatureRegistryProtocol(Protocol):
    def get_feature_sets(self) -> Mapping[str, Type[Features]]: ...

    def add_feature_set(self, updated_class: Type[Features]): ...

    def get_singletons(self) -> Mapping[str, Type[Features]]: ...


class FeatureSetBase:
    """Registry containing all @features classes."""

    registry: ClassVar[Dict[str, Type[Features]]] = {}
    """Mapping of fqn to Features cls"""

    __chalk_singletons__: ClassVar[Dict[str, Type[Features]]] = {}
    """Mapping of fqn to singleton Features cls. These are also included in `.registry`."""

    __chalk_notebook_defined_feature_fields__: ClassVar[Dict[str, Set[str]]] = defaultdict(set)
    """
    (Mapping of namespace to a set of unversioned attribute names).
    This tracks which feature fields have been added to a feature class in a notebook, after it's been defined.
    """

    hook: ClassVar[Optional[Callable[[Type[Features]], None]]] = None

    def __init__(self) -> None:
        super().__init__()
        raise RuntimeError("FeatureSetBase should never be instantiated")


class FeatureSetBaseWrapper(FeatureRegistryProtocol):
    """
    Wrapper around the global FeatureSetBase
    """

    def __init__(self):
        super().__init__()

    def get_feature_sets(self) -> Mapping[str, Type[Features]]:
        return FeatureSetBase.registry

    def add_feature_set(self, updated_class: Type[Features]):
        namespace = updated_class.namespace
        previous_features_class = FeatureSetBase.registry.get(namespace)
        if (
            notebook.is_notebook()
            and previous_features_class is not None
            and notebook.is_defined_in_module(previous_features_class)
        ):
            # Not generating an LSP here because we're in a notebook anyway
            # TODO: See if we can pretty-print lsp errors in notebooks, at which point we can generate one that points to the old feature class
            raise ValueError(
                f"Cannot re-define feature class '{previous_features_class.__name__}' in a notebook: it was previously defined in '{previous_features_class.__module__}'."
            )

        FeatureSetBase.registry[updated_class.__chalk_namespace__] = updated_class

        if updated_class.__chalk_is_singleton__:
            FeatureSetBase.__chalk_singletons__[updated_class.__chalk_namespace__] = updated_class

        if FeatureSetBase.hook is not None:
            try:
                FeatureSetBase.hook(updated_class)
            except:
                # If we were unsuccessful, restore the previous features class, and call the hook again
                if previous_features_class is not None:
                    FeatureSetBase.registry[updated_class.__chalk_namespace__] = previous_features_class
                    if updated_class.__chalk_is_singleton__:
                        FeatureSetBase.__chalk_singletons__[updated_class.__chalk_namespace__] = previous_features_class
                    FeatureSetBase.hook(previous_features_class)
                # Now raise the original exception
                raise

        ...

    def get_singletons(self) -> Mapping[str, Type[Features]]:
        return FeatureSetBase.__chalk_singletons__


_DEFAULT_FEATURE_REGISTRY: FeatureRegistryProtocol = FeatureSetBaseWrapper()
CURRENT_FEATURE_REGISTRY: contextvars.ContextVar[FeatureRegistryProtocol] = contextvars.ContextVar(
    "CURRENT_FEATURE_REGISTRY", default=_DEFAULT_FEATURE_REGISTRY
)


def is_features_cls(maybe_features: Any) -> TypeGuard[Type[Features]]:
    """BEWARE: this returns true for FeatureClass but also Features[...]"""
    return isinstance(maybe_features, type) and issubclass(maybe_features, Features)


def is_feature_set_class(maybe_features: Any) -> TypeGuard[Type[FeaturesImpl]]:
    return is_features_cls(maybe_features) and getattr(maybe_features, "__chalk_feature_set__", False)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/auth.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import audit_pb2 as chalk_dot_auth_dot_v1_dot_audit__pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from chalk._gen.chalk.server.v1 import link_pb2 as chalk_dot_server_dot_v1_dot_link__pb2
from chalk._gen.chalk.utils.v1 import sensitive_pb2 as chalk_dot_utils_dot_v1_dot_sensitive__pb2
from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1a\x63halk/server/v1/auth.proto\x12\x0f\x63halk.server.v1\x1a\x19\x63halk/auth/v1/audit.proto\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a\x1a\x63halk/server/v1/link.proto\x1a\x1e\x63halk/utils/v1/sensitive.proto\x1a google/protobuf/descriptor.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\x9d\x01\n\x0fGetTokenRequest\x12\x1b\n\tclient_id\x18\x01 \x01(\tR\x08\x63lientId\x12)\n\rclient_secret\x18\x02 \x01(\tB\x04\xd8\xa1\'\x01R\x0c\x63lientSecret\x12\x1d\n\ngrant_type\x18\x03 \x01(\tR\tgrantType\x12\x19\n\x05scope\x18\x04 \x01(\tH\x00R\x05scope\x88\x01\x01\x42\x08\n\x06_scope"\xf7\x05\n\x10GetTokenResponse\x12\'\n\x0c\x61\x63\x63\x65ss_token\x18\x01 \x01(\tB\x04\xd8\xa1\'\x01R\x0b\x61\x63\x63\x65ssToken\x12\x1d\n\ntoken_type\x18\x02 \x01(\tR\ttokenType\x12\x1d\n\nexpires_in\x18\x03 \x01(\x05R\texpiresIn\x12\x39\n\nexpires_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\texpiresAt\x12\x1d\n\napi_server\x18\x05 \x01(\tR\tapiServer\x12\x34\n\x13primary_environment\x18\x06 \x01(\tH\x00R\x12primaryEnvironment\x88\x01\x01\x12H\n\x07\x65ngines\x18\x07 \x03(\x0b\x32..chalk.server.v1.GetTokenResponse.EnginesEntryR\x07\x65ngines\x12U\n\x0cgrpc_engines\x18\x08 \x03(\x0b\x32\x32.chalk.server.v1.GetTokenResponse.GrpcEnginesEntryR\x0bgrpcEngines\x12o\n\x16\x65nvironment_id_to_name\x18\t \x03(\x0b\x32:.chalk.server.v1.GetTokenResponse.EnvironmentIdToNameEntryR\x13\x65nvironmentIdToName\x1a:\n\x0c\x45nginesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a>\n\x10GrpcEnginesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a\x46\n\x18\x45nvironmentIdToNameEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x16\n\x14_primary_environment"j\n\x18UpdateLinkSessionRequest\x12\x16\n\x06status\x18\x01 \x01(\tR\x06status\x12\x17\n\x07user_id\x18\x02 \x01(\tR\x06userId\x12\x1d\n\nsession_id\x18\x03 \x01(\tR\tsessionId"+\n\x19UpdateLinkSessionResponse\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id2\xbf\x03\n\x0b\x41uthService\x12Z\n\x08GetToken\x12 .chalk.server.v1.GetTokenRequest\x1a!.chalk.server.v1.GetTokenResponse"\t\x80}\x01\x8a\xd3\x0e\x02\x08\x01\x12u\n\x11\x43reateLinkSession\x12).chalk.server.v1.CreateLinkSessionRequest\x1a*.chalk.server.v1.CreateLinkSessionResponse"\t\x80}\x01\x8a\xd3\x0e\x02\x08\x01\x12\x66\n\x0eGetLinkSession\x12&.chalk.server.v1.GetLinkSessionRequest\x1a\'.chalk.server.v1.GetLinkSessionResponse"\x03\x80}\x01\x12u\n\x11UpdateLinkSession\x12).chalk.server.v1.UpdateLinkSessionRequest\x1a*.chalk.server.v1.UpdateLinkSessionResponse"\t\x80}\x01\x8a\xd3\x0e\x02\x08\x01\x42\x92\x01\n\x13\x63om.chalk.server.v1B\tAuthProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.auth_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\tAuthProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_GETTOKENREQUEST"].fields_by_name["client_secret"]._options = None
    _globals["_GETTOKENREQUEST"].fields_by_name["client_secret"]._serialized_options = b"\330\241'\001"
    _globals["_GETTOKENRESPONSE_ENGINESENTRY"]._options = None
    _globals["_GETTOKENRESPONSE_ENGINESENTRY"]._serialized_options = b"8\001"
    _globals["_GETTOKENRESPONSE_GRPCENGINESENTRY"]._options = None
    _globals["_GETTOKENRESPONSE_GRPCENGINESENTRY"]._serialized_options = b"8\001"
    _globals["_GETTOKENRESPONSE_ENVIRONMENTIDTONAMEENTRY"]._options = None
    _globals["_GETTOKENRESPONSE_ENVIRONMENTIDTONAMEENTRY"]._serialized_options = b"8\001"
    _globals["_GETTOKENRESPONSE"].fields_by_name["access_token"]._options = None
    _globals["_GETTOKENRESPONSE"].fields_by_name["access_token"]._serialized_options = b"\330\241'\001"
    _globals["_AUTHSERVICE"].methods_by_name["GetToken"]._options = None
    _globals["_AUTHSERVICE"].methods_by_name["GetToken"]._serialized_options = b"\200}\001\212\323\016\002\010\001"
    _globals["_AUTHSERVICE"].methods_by_name["CreateLinkSession"]._options = None
    _globals["_AUTHSERVICE"].methods_by_name[
        "CreateLinkSession"
    ]._serialized_options = b"\200}\001\212\323\016\002\010\001"
    _globals["_AUTHSERVICE"].methods_by_name["GetLinkSession"]._options = None
    _globals["_AUTHSERVICE"].methods_by_name["GetLinkSession"]._serialized_options = b"\200}\001"
    _globals["_AUTHSERVICE"].methods_by_name["UpdateLinkSession"]._options = None
    _globals["_AUTHSERVICE"].methods_by_name[
        "UpdateLinkSession"
    ]._serialized_options = b"\200}\001\212\323\016\002\010\001"
    _globals["_GETTOKENREQUEST"]._serialized_start = 235
    _globals["_GETTOKENREQUEST"]._serialized_end = 392
    _globals["_GETTOKENRESPONSE"]._serialized_start = 395
    _globals["_GETTOKENRESPONSE"]._serialized_end = 1154
    _globals["_GETTOKENRESPONSE_ENGINESENTRY"]._serialized_start = 936
    _globals["_GETTOKENRESPONSE_ENGINESENTRY"]._serialized_end = 994
    _globals["_GETTOKENRESPONSE_GRPCENGINESENTRY"]._serialized_start = 996
    _globals["_GETTOKENRESPONSE_GRPCENGINESENTRY"]._serialized_end = 1058
    _globals["_GETTOKENRESPONSE_ENVIRONMENTIDTONAMEENTRY"]._serialized_start = 1060
    _globals["_GETTOKENRESPONSE_ENVIRONMENTIDTONAMEENTRY"]._serialized_end = 1130
    _globals["_UPDATELINKSESSIONREQUEST"]._serialized_start = 1156
    _globals["_UPDATELINKSESSIONREQUEST"]._serialized_end = 1262
    _globals["_UPDATELINKSESSIONRESPONSE"]._serialized_start = 1264
    _globals["_UPDATELINKSESSIONRESPONSE"]._serialized_end = 1307
    _globals["_AUTHSERVICE"]._serialized_start = 1310
    _globals["_AUTHSERVICE"]._serialized_end = 1757
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.queries_pb2 import (
    GetQueryErrorsChartRequest,
    GetQueryErrorsChartResponse,
    GetQueryPerformanceSummaryRequest,
    GetQueryPerformanceSummaryResponse,
    GetQueryPlanRequest,
    GetQueryPlanResponse,
    ListQueryErrorsRequest,
    ListQueryErrorsResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class QueriesServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetQueryPerformanceSummary: UnaryUnaryMultiCallable[
        GetQueryPerformanceSummaryRequest,
        GetQueryPerformanceSummaryResponse,
    ]
    ListQueryErrors: UnaryUnaryMultiCallable[
        ListQueryErrorsRequest,
        ListQueryErrorsResponse,
    ]
    GetQueryErrorsChart: UnaryUnaryMultiCallable[
        GetQueryErrorsChartRequest,
        GetQueryErrorsChartResponse,
    ]
    GetQueryPlan: UnaryUnaryMultiCallable[
        GetQueryPlanRequest,
        GetQueryPlanResponse,
    ]

class QueriesServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetQueryPerformanceSummary(
        self,
        request: GetQueryPerformanceSummaryRequest,
        context: ServicerContext,
    ) -> GetQueryPerformanceSummaryResponse: ...
    @abstractmethod
    def ListQueryErrors(
        self,
        request: ListQueryErrorsRequest,
        context: ServicerContext,
    ) -> ListQueryErrorsResponse: ...
    @abstractmethod
    def GetQueryErrorsChart(
        self,
        request: GetQueryErrorsChartRequest,
        context: ServicerContext,
    ) -> GetQueryErrorsChartResponse: ...
    @abstractmethod
    def GetQueryPlan(
        self,
        request: GetQueryPlanRequest,
        context: ServicerContext,
    ) -> GetQueryPlanResponse: ...

def add_QueriesServiceServicer_to_server(servicer: QueriesServiceServicer, server: Server) -> None: ...



from chalk._gen.chalk.arrow.v1 import arrow_pb2 as _arrow_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class RawColumnKey(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class FeatureColumnKey(_message.Message):
    __slots__ = ("root_fqn",)
    ROOT_FQN_FIELD_NUMBER: _ClassVar[int]
    root_fqn: str
    def __init__(self, root_fqn: _Optional[str] = ...) -> None: ...

class HasManyFeatureKey(_message.Message):
    __slots__ = ("root_fqn", "df")
    ROOT_FQN_FIELD_NUMBER: _ClassVar[int]
    DF_FIELD_NUMBER: _ClassVar[int]
    root_fqn: str
    df: DataFrame
    def __init__(self, root_fqn: _Optional[str] = ..., df: _Optional[_Union[DataFrame, _Mapping]] = ...) -> None: ...

class DataFrameFeatureKey(_message.Message):
    __slots__ = ("root_namespace", "df")
    ROOT_NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    DF_FIELD_NUMBER: _ClassVar[int]
    root_namespace: str
    df: DataFrame
    def __init__(
        self, root_namespace: _Optional[str] = ..., df: _Optional[_Union[DataFrame, _Mapping]] = ...
    ) -> None: ...

class DataFrame(_message.Message):
    __slots__ = ("optional_columns", "required_columns", "limit")
    OPTIONAL_COLUMNS_FIELD_NUMBER: _ClassVar[int]
    REQUIRED_COLUMNS_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    optional_columns: _containers.RepeatedCompositeFieldContainer[FeatureColumnKey]
    required_columns: _containers.RepeatedCompositeFieldContainer[FeatureColumnKey]
    limit: int
    def __init__(
        self,
        optional_columns: _Optional[_Iterable[_Union[FeatureColumnKey, _Mapping]]] = ...,
        required_columns: _Optional[_Iterable[_Union[FeatureColumnKey, _Mapping]]] = ...,
        limit: _Optional[int] = ...,
    ) -> None: ...

class ColumnKey(_message.Message):
    __slots__ = ("feature", "raw")
    FEATURE_FIELD_NUMBER: _ClassVar[int]
    RAW_FIELD_NUMBER: _ClassVar[int]
    feature: FeatureColumnKey
    raw: RawColumnKey
    def __init__(
        self,
        feature: _Optional[_Union[FeatureColumnKey, _Mapping]] = ...,
        raw: _Optional[_Union[RawColumnKey, _Mapping]] = ...,
    ) -> None: ...

class TableKey(_message.Message):
    __slots__ = ("has_many", "data_frame")
    HAS_MANY_FIELD_NUMBER: _ClassVar[int]
    DATA_FRAME_FIELD_NUMBER: _ClassVar[int]
    has_many: HasManyFeatureKey
    data_frame: DataFrameFeatureKey
    def __init__(
        self,
        has_many: _Optional[_Union[HasManyFeatureKey, _Mapping]] = ...,
        data_frame: _Optional[_Union[DataFrameFeatureKey, _Mapping]] = ...,
    ) -> None: ...

class Key(_message.Message):
    __slots__ = ("raw_column", "scalar", "has_many", "dataframe")
    RAW_COLUMN_FIELD_NUMBER: _ClassVar[int]
    SCALAR_FIELD_NUMBER: _ClassVar[int]
    HAS_MANY_FIELD_NUMBER: _ClassVar[int]
    DATAFRAME_FIELD_NUMBER: _ClassVar[int]
    raw_column: RawColumnKey
    scalar: FeatureColumnKey
    has_many: HasManyFeatureKey
    dataframe: DataFrameFeatureKey
    def __init__(
        self,
        raw_column: _Optional[_Union[RawColumnKey, _Mapping]] = ...,
        scalar: _Optional[_Union[FeatureColumnKey, _Mapping]] = ...,
        has_many: _Optional[_Union[HasManyFeatureKey, _Mapping]] = ...,
        dataframe: _Optional[_Union[DataFrameFeatureKey, _Mapping]] = ...,
    ) -> None: ...

class PyArrowSchema(_message.Message):
    __slots__ = ("scalars", "groups")
    class TableSchema(_message.Message):
        __slots__ = ("key", "schema")
        class SchemaEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: _arrow_pb2.ArrowType
            def __init__(
                self, key: _Optional[str] = ..., value: _Optional[_Union[_arrow_pb2.ArrowType, _Mapping]] = ...
            ) -> None: ...

        KEY_FIELD_NUMBER: _ClassVar[int]
        SCHEMA_FIELD_NUMBER: _ClassVar[int]
        key: TableKey
        schema: _containers.MessageMap[str, _arrow_pb2.ArrowType]
        def __init__(
            self,
            key: _Optional[_Union[TableKey, _Mapping]] = ...,
            schema: _Optional[_Mapping[str, _arrow_pb2.ArrowType]] = ...,
        ) -> None: ...

    class ColumnSchema(_message.Message):
        __slots__ = ("key", "schema")
        KEY_FIELD_NUMBER: _ClassVar[int]
        SCHEMA_FIELD_NUMBER: _ClassVar[int]
        key: ColumnKey
        schema: _arrow_pb2.ArrowType
        def __init__(
            self,
            key: _Optional[_Union[ColumnKey, _Mapping]] = ...,
            schema: _Optional[_Union[_arrow_pb2.ArrowType, _Mapping]] = ...,
        ) -> None: ...

    SCALARS_FIELD_NUMBER: _ClassVar[int]
    GROUPS_FIELD_NUMBER: _ClassVar[int]
    scalars: _containers.RepeatedCompositeFieldContainer[PyArrowSchema.ColumnSchema]
    groups: _containers.RepeatedCompositeFieldContainer[PyArrowSchema.TableSchema]
    def __init__(
        self,
        scalars: _Optional[_Iterable[_Union[PyArrowSchema.ColumnSchema, _Mapping]]] = ...,
        groups: _Optional[_Iterable[_Union[PyArrowSchema.TableSchema, _Mapping]]] = ...,
    ) -> None: ...

class Plan(_message.Message):
    __slots__ = ("nodes", "root_node_idx", "pyarrow_schema")
    NODES_FIELD_NUMBER: _ClassVar[int]
    ROOT_NODE_IDX_FIELD_NUMBER: _ClassVar[int]
    PYARROW_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    nodes: _containers.RepeatedCompositeFieldContainer[Node]
    root_node_idx: int
    pyarrow_schema: PyArrowSchema
    def __init__(
        self,
        nodes: _Optional[_Iterable[_Union[Node, _Mapping]]] = ...,
        root_node_idx: _Optional[int] = ...,
        pyarrow_schema: _Optional[_Union[PyArrowSchema, _Mapping]] = ...,
    ) -> None: ...

class Node(_message.Message):
    __slots__ = ("children_indices", "impl")
    CHILDREN_INDICES_FIELD_NUMBER: _ClassVar[int]
    IMPL_FIELD_NUMBER: _ClassVar[int]
    children_indices: _containers.RepeatedScalarFieldContainer[int]
    impl: NodeImpl
    def __init__(
        self, children_indices: _Optional[_Iterable[int]] = ..., impl: _Optional[_Union[NodeImpl, _Mapping]] = ...
    ) -> None: ...

class NodeImpl(_message.Message):
    __slots__ = ("unknown", "givens_scan", "project", "chalk_project", "default_injector")
    UNKNOWN_FIELD_NUMBER: _ClassVar[int]
    GIVENS_SCAN_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    CHALK_PROJECT_FIELD_NUMBER: _ClassVar[int]
    DEFAULT_INJECTOR_FIELD_NUMBER: _ClassVar[int]
    unknown: UnknownNode
    givens_scan: GivensScan
    project: Project
    chalk_project: ChalkProject
    default_injector: DefaultInjector
    def __init__(
        self,
        unknown: _Optional[_Union[UnknownNode, _Mapping]] = ...,
        givens_scan: _Optional[_Union[GivensScan, _Mapping]] = ...,
        project: _Optional[_Union[Project, _Mapping]] = ...,
        chalk_project: _Optional[_Union[ChalkProject, _Mapping]] = ...,
        default_injector: _Optional[_Union[DefaultInjector, _Mapping]] = ...,
    ) -> None: ...

class UnknownNode(_message.Message):
    __slots__ = ("type_name",)
    TYPE_NAME_FIELD_NUMBER: _ClassVar[int]
    type_name: str
    def __init__(self, type_name: _Optional[str] = ...) -> None: ...

class GivensScan(_message.Message):
    __slots__ = ("fields",)
    FIELDS_FIELD_NUMBER: _ClassVar[int]
    fields: _containers.RepeatedCompositeFieldContainer[Key]
    def __init__(self, fields: _Optional[_Iterable[_Union[Key, _Mapping]]] = ...) -> None: ...

class ChalkProject(_message.Message):
    __slots__ = ("fields", "promote_ts_to_feature_time")
    FIELDS_FIELD_NUMBER: _ClassVar[int]
    PROMOTE_TS_TO_FEATURE_TIME_FIELD_NUMBER: _ClassVar[int]
    fields: _containers.RepeatedCompositeFieldContainer[Key]
    promote_ts_to_feature_time: bool
    def __init__(
        self, fields: _Optional[_Iterable[_Union[Key, _Mapping]]] = ..., promote_ts_to_feature_time: bool = ...
    ) -> None: ...

class Project(_message.Message):
    __slots__ = ("fields",)
    FIELDS_FIELD_NUMBER: _ClassVar[int]
    fields: _containers.RepeatedCompositeFieldContainer[Key]
    def __init__(self, fields: _Optional[_Iterable[_Union[Key, _Mapping]]] = ...) -> None: ...

class DefaultInjector(_message.Message):
    __slots__ = ("defaulting_features",)
    DEFAULTING_FEATURES_FIELD_NUMBER: _ClassVar[int]
    defaulting_features: _containers.RepeatedCompositeFieldContainer[Key]
    def __init__(self, defaulting_features: _Optional[_Iterable[_Union[Key, _Mapping]]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/topicpush.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from chalk._gen.chalk.server.v1 import topic_pb2 as chalk_dot_server_dot_v1_dot_topic__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1f\x63halk/server/v1/topicpush.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a\x1b\x63halk/server/v1/topic.proto"\xef\x02\n\x0cScheduledJob\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x18\n\x07payload\x18\x03 \x01(\tR\x07payload\x12\x1a\n\x08schedule\x18\x04 \x01(\tR\x08schedule\x12,\n\x05topic\x18\x05 \x01(\x0b\x32\x16.chalk.server.v1.TopicR\x05topic\x12M\n\nattributes\x18\x06 \x03(\x0b\x32-.chalk.server.v1.ScheduledJob.AttributesEntryR\nattributes\x12%\n\x0b\x65nvironment\x18\x07 \x01(\tH\x00R\x0b\x65nvironment\x88\x01\x01\x12\x12\n\x04tags\x18\x08 \x03(\tR\x04tags\x1a=\n\x0f\x41ttributesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x0e\n\x0c_environment"\xf7\x02\n\x10UpdateJobRequest\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x18\n\x07payload\x18\x03 \x01(\tR\x07payload\x12\x1a\n\x08schedule\x18\x04 \x01(\tR\x08schedule\x12,\n\x05topic\x18\x05 \x01(\x0b\x32\x16.chalk.server.v1.TopicR\x05topic\x12Q\n\nattributes\x18\x06 \x03(\x0b\x32\x31.chalk.server.v1.UpdateJobRequest.AttributesEntryR\nattributes\x12%\n\x0b\x65nvironment\x18\x07 \x01(\tH\x00R\x0b\x65nvironment\x88\x01\x01\x12\x12\n\x04tags\x18\x08 \x03(\tR\x04tags\x1a=\n\x0f\x41ttributesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x0e\n\x0c_environment"D\n\x11UpdateJobResponse\x12/\n\x03job\x18\x01 \x01(\x0b\x32\x1d.chalk.server.v1.ScheduledJobR\x03job"\x11\n\x0fListJobsRequest"E\n\x10ListJobsResponse\x12\x31\n\x04jobs\x18\x01 \x03(\x0b\x32\x1d.chalk.server.v1.ScheduledJobR\x04jobs"\xe7\x02\n\x10\x43reateJobRequest\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x18\n\x07payload\x18\x02 \x01(\tR\x07payload\x12\x1a\n\x08schedule\x18\x03 \x01(\tR\x08schedule\x12,\n\x05topic\x18\x04 \x01(\x0b\x32\x16.chalk.server.v1.TopicR\x05topic\x12Q\n\nattributes\x18\x05 \x03(\x0b\x32\x31.chalk.server.v1.CreateJobRequest.AttributesEntryR\nattributes\x12%\n\x0b\x65nvironment\x18\x07 \x01(\tH\x00R\x0b\x65nvironment\x88\x01\x01\x12\x12\n\x04tags\x18\x08 \x03(\tR\x04tags\x1a=\n\x0f\x41ttributesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x0e\n\x0c_environment"D\n\x11\x43reateJobResponse\x12/\n\x03job\x18\x01 \x01(\x0b\x32\x1d.chalk.server.v1.ScheduledJobR\x03job""\n\x10\x44\x65leteJobRequest\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id"\x13\n\x11\x44\x65leteJobResponse")\n\x13GetJobByNameRequest\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name"G\n\x14GetJobByNameResponse\x12/\n\x03job\x18\x01 \x01(\x0b\x32\x1d.chalk.server.v1.ScheduledJobR\x03job2\xde\x03\n\x10TopicPushService\x12W\n\x08ListJobs\x12 .chalk.server.v1.ListJobsRequest\x1a!.chalk.server.v1.ListJobsResponse"\x06\x90\x02\x01\x80}\x1b\x12W\n\tCreateJob\x12!.chalk.server.v1.CreateJobRequest\x1a".chalk.server.v1.CreateJobResponse"\x03\x80}\x1b\x12Z\n\tUpdateJob\x12!.chalk.server.v1.UpdateJobRequest\x1a".chalk.server.v1.UpdateJobResponse"\x06\x90\x02\x02\x80}\x1b\x12W\n\tDeleteJob\x12!.chalk.server.v1.DeleteJobRequest\x1a".chalk.server.v1.DeleteJobResponse"\x03\x80}\x1b\x12\x63\n\x0cGetJobByName\x12$.chalk.server.v1.GetJobByNameRequest\x1a%.chalk.server.v1.GetJobByNameResponse"\x06\x90\x02\x02\x80}\x1b\x42\x97\x01\n\x13\x63om.chalk.server.v1B\x0eTopicpushProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.topicpush_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\016TopicpushProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_SCHEDULEDJOB_ATTRIBUTESENTRY"]._options = None
    _globals["_SCHEDULEDJOB_ATTRIBUTESENTRY"]._serialized_options = b"8\001"
    _globals["_UPDATEJOBREQUEST_ATTRIBUTESENTRY"]._options = None
    _globals["_UPDATEJOBREQUEST_ATTRIBUTESENTRY"]._serialized_options = b"8\001"
    _globals["_CREATEJOBREQUEST_ATTRIBUTESENTRY"]._options = None
    _globals["_CREATEJOBREQUEST_ATTRIBUTESENTRY"]._serialized_options = b"8\001"
    _globals["_TOPICPUSHSERVICE"].methods_by_name["ListJobs"]._options = None
    _globals["_TOPICPUSHSERVICE"].methods_by_name["ListJobs"]._serialized_options = b"\220\002\001\200}\033"
    _globals["_TOPICPUSHSERVICE"].methods_by_name["CreateJob"]._options = None
    _globals["_TOPICPUSHSERVICE"].methods_by_name["CreateJob"]._serialized_options = b"\200}\033"
    _globals["_TOPICPUSHSERVICE"].methods_by_name["UpdateJob"]._options = None
    _globals["_TOPICPUSHSERVICE"].methods_by_name["UpdateJob"]._serialized_options = b"\220\002\002\200}\033"
    _globals["_TOPICPUSHSERVICE"].methods_by_name["DeleteJob"]._options = None
    _globals["_TOPICPUSHSERVICE"].methods_by_name["DeleteJob"]._serialized_options = b"\200}\033"
    _globals["_TOPICPUSHSERVICE"].methods_by_name["GetJobByName"]._options = None
    _globals["_TOPICPUSHSERVICE"].methods_by_name["GetJobByName"]._serialized_options = b"\220\002\002\200}\033"
    _globals["_SCHEDULEDJOB"]._serialized_start = 115
    _globals["_SCHEDULEDJOB"]._serialized_end = 482
    _globals["_SCHEDULEDJOB_ATTRIBUTESENTRY"]._serialized_start = 405
    _globals["_SCHEDULEDJOB_ATTRIBUTESENTRY"]._serialized_end = 466
    _globals["_UPDATEJOBREQUEST"]._serialized_start = 485
    _globals["_UPDATEJOBREQUEST"]._serialized_end = 860
    _globals["_UPDATEJOBREQUEST_ATTRIBUTESENTRY"]._serialized_start = 405
    _globals["_UPDATEJOBREQUEST_ATTRIBUTESENTRY"]._serialized_end = 466
    _globals["_UPDATEJOBRESPONSE"]._serialized_start = 862
    _globals["_UPDATEJOBRESPONSE"]._serialized_end = 930
    _globals["_LISTJOBSREQUEST"]._serialized_start = 932
    _globals["_LISTJOBSREQUEST"]._serialized_end = 949
    _globals["_LISTJOBSRESPONSE"]._serialized_start = 951
    _globals["_LISTJOBSRESPONSE"]._serialized_end = 1020
    _globals["_CREATEJOBREQUEST"]._serialized_start = 1023
    _globals["_CREATEJOBREQUEST"]._serialized_end = 1382
    _globals["_CREATEJOBREQUEST_ATTRIBUTESENTRY"]._serialized_start = 405
    _globals["_CREATEJOBREQUEST_ATTRIBUTESENTRY"]._serialized_end = 466
    _globals["_CREATEJOBRESPONSE"]._serialized_start = 1384
    _globals["_CREATEJOBRESPONSE"]._serialized_end = 1452
    _globals["_DELETEJOBREQUEST"]._serialized_start = 1454
    _globals["_DELETEJOBREQUEST"]._serialized_end = 1488
    _globals["_DELETEJOBRESPONSE"]._serialized_start = 1490
    _globals["_DELETEJOBRESPONSE"]._serialized_end = 1509
    _globals["_GETJOBBYNAMEREQUEST"]._serialized_start = 1511
    _globals["_GETJOBBYNAMEREQUEST"]._serialized_end = 1552
    _globals["_GETJOBBYNAMERESPONSE"]._serialized_start = 1554
    _globals["_GETJOBBYNAMERESPONSE"]._serialized_end = 1625
    _globals["_TOPICPUSHSERVICE"]._serialized_start = 1628
    _globals["_TOPICPUSHSERVICE"]._serialized_end = 2106
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Union

from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.sql._internal.sql_source import BaseSQLSource, SQLSourceKind
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from sqlalchemy.engine.url import URL

_DATABRICKS_HOST_NAME = "DATABRICKS_HOST"
_DATABRICKS_HTTP_PATH_NAME = "DATABRICKS_HTTP_PATH"
_DATABRICKS_TOKEN_NAME = "DATABRICKS_TOKEN"
_DATABRICKS_DATABASE_NAME = "DATABRICKS_DATABASE"
_DATABRICKS_PORT_NAME = "DATABRICKS_PORT"


class DatabricksSourceImpl(BaseSQLSource):
    kind = SQLSourceKind.databricks

    def __init__(
        self,
        host: Optional[str] = None,
        http_path: Optional[str] = None,
        access_token: Optional[str] = None,
        db: Optional[str] = None,
        port: Optional[Union[int, str]] = None,
        name: Optional[str] = None,
        engine_args: Optional[Dict[str, Any]] = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        try:
            from databricks import sql
        except ImportError:
            raise missing_dependency_exception("chalkpy[databricks]")
        del sql
        self.host = host or load_integration_variable(
            name=_DATABRICKS_HOST_NAME, integration_name=name, override=integration_variable_override
        )
        self.http_path = http_path or load_integration_variable(
            name=_DATABRICKS_HTTP_PATH_NAME, integration_name=name, override=integration_variable_override
        )
        self.access_token = access_token or load_integration_variable(
            name=_DATABRICKS_TOKEN_NAME, integration_name=name, override=integration_variable_override
        )
        self.db = db or load_integration_variable(
            name=_DATABRICKS_DATABASE_NAME, integration_name=name, override=integration_variable_override
        )
        self.port = (
            int(port)
            if port is not None
            else load_integration_variable(
                name=_DATABRICKS_PORT_NAME, integration_name=name, parser=int, override=integration_variable_override
            )
        )
        if engine_args is None:
            engine_args = {}
        engine_args.setdefault("pool_size", 20)
        engine_args.setdefault("max_overflow", 60)
        engine_args.setdefault(
            "connect_args",
            {
                "keepalives": 1,
                "keepalives_idle": 30,
                "keepalives_interval": 10,
                "keepalives_count": 5,
            },
        )
        BaseSQLSource.__init__(self, name=name, engine_args=engine_args, async_engine_args={})

    def get_sqlglot_dialect(self) -> str | None:
        return "databricks"

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="databricks",
            username="token",
            password=self.access_token,
            host=self.host,
            port=self.port,
            database=self.db,
            query={"http_path": self.http_path or ""},
        )

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(_DATABRICKS_HOST_NAME, self.name, self.host),
                create_integration_variable(_DATABRICKS_HTTP_PATH_NAME, self.name, self.http_path),
                create_integration_variable(_DATABRICKS_TOKEN_NAME, self.name, self.access_token),
                create_integration_variable(_DATABRICKS_DATABASE_NAME, self.name, self.db),
                create_integration_variable(_DATABRICKS_PORT_NAME, self.name, self.port),
            ]
            if v is not None
        }



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/billing.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from chalk._gen.chalk.kubernetes.v1 import nodes_pb2 as chalk_dot_kubernetes_dot_v1_dot_nodes__pb2
from chalk._gen.chalk.kubernetes.v1 import pods_pb2 as chalk_dot_kubernetes_dot_v1_dot_pods__pb2
from chalk._gen.chalk.pubsub.v1 import node_status_pb2 as chalk_dot_pubsub_dot_v1_dot_node__status__pb2
from chalk._gen.chalk.pubsub.v1 import pod_status_pb2 as chalk_dot_pubsub_dot_v1_dot_pod__status__pb2
from chalk._gen.chalk.server.v1 import chart_pb2 as chalk_dot_server_dot_v1_dot_chart__pb2
from chalk._gen.chalk.server.v1 import pod_request_pb2 as chalk_dot_server_dot_v1_dot_pod__request__pb2
from chalk._gen.chalk.usage.v1 import rate_pb2 as chalk_dot_usage_dot_v1_dot_rate__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1d\x63halk/server/v1/billing.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a\x1f\x63halk/kubernetes/v1/nodes.proto\x1a\x1e\x63halk/kubernetes/v1/pods.proto\x1a!chalk/pubsub/v1/node_status.proto\x1a chalk/pubsub/v1/pod_status.proto\x1a\x1b\x63halk/server/v1/chart.proto\x1a!chalk/server/v1/pod_request.proto\x1a\x19\x63halk/usage/v1/rate.proto"\x88\x02\n\x14GetUsageChartRequest\x12\x1e\n\x08start_ms\x18\x01 \x01(\x03H\x00R\x07startMs\x88\x01\x01\x12\x1a\n\x06\x65nd_ms\x18\x02 \x01(\x03H\x01R\x05\x65ndMs\x88\x01\x01\x12>\n\x06period\x18\x03 \x01(\x0e\x32!.chalk.server.v1.UsageChartPeriodH\x02R\x06period\x88\x01\x01\x12\x44\n\x08grouping\x18\x04 \x01(\x0e\x32#.chalk.server.v1.UsageChartGroupingH\x03R\x08grouping\x88\x01\x01\x42\x0b\n\t_start_msB\t\n\x07_end_msB\t\n\x07_periodB\x0b\n\t_grouping"E\n\x15GetUsageChartResponse\x12,\n\x05\x63hart\x18\x01 \x01(\x0b\x32\x16.chalk.server.v1.ChartR\x05\x63hart"\x1c\n\x1aGetUtilizationRatesRequest"P\n\x1bGetUtilizationRatesResponse\x12\x31\n\x05rates\x18\x01 \x03(\x0b\x32\x1b.chalk.usage.v1.MachineRateR\x05rates"\xd2\x01\n\x16GetNodesAndPodsRequest\x12!\n\tnamespace\x18\x01 \x01(\tH\x00R\tnamespace\x88\x01\x01\x12\x31\n\x12pod_label_selector\x18\x02 \x01(\tH\x01R\x10podLabelSelector\x88\x01\x01\x12*\n\x0e\x65nvironment_id\x18\x03 \x01(\tH\x02R\renvironmentId\x88\x01\x01\x42\x0c\n\n_namespaceB\x15\n\x13_pod_label_selectorB\x11\n\x0f_environment_id"\x88\x01\n\x17GetNodesAndPodsResponse\x12\x37\n\x05nodes\x18\x01 \x03(\x0b\x32!.chalk.pubsub.v1.NodeStatusPubSubR\x05nodes\x12\x34\n\x04pods\x18\x02 \x03(\x0b\x32 .chalk.pubsub.v1.PodStatusPubSubR\x04pods"\xd4\x01\n\x18GetNodesAndPodsUIRequest\x12!\n\tnamespace\x18\x01 \x01(\tH\x00R\tnamespace\x88\x01\x01\x12\x31\n\x12pod_label_selector\x18\x02 \x01(\tH\x01R\x10podLabelSelector\x88\x01\x01\x12*\n\x0e\x65nvironment_id\x18\x03 \x01(\tH\x02R\renvironmentId\x88\x01\x01\x42\x0c\n\n_namespaceB\x15\n\x13_pod_label_selectorB\x11\n\x0f_environment_id"\x96\x01\n\x19GetNodesAndPodsUIResponse\x12=\n\x05nodes\x18\x01 \x03(\x0b\x32\'.chalk.kubernetes.v1.KubernetesNodeDataR\x05nodes\x12:\n\x04pods\x18\x02 \x03(\x0b\x32&.chalk.kubernetes.v1.KubernetesPodDataR\x04pods"\x18\n\x16SyncUtilizationRequest"\x19\n\x17SyncUtilizationResponse*t\n\x10UsageChartPeriod\x12"\n\x1eUSAGE_CHART_PERIOD_UNSPECIFIED\x10\x00\x12\x1c\n\x18USAGE_CHART_PERIOD_DAILY\x10\x01\x12\x1e\n\x1aUSAGE_CHART_PERIOD_MONTHLY\x10\x02*\x84\x01\n\x12UsageChartGrouping\x12$\n USAGE_CHART_GROUPING_UNSPECIFIED\x10\x00\x12&\n"USAGE_CHART_GROUPING_INSTANCE_TYPE\x10\x01\x12 \n\x1cUSAGE_CHART_GROUPING_CLUSTER\x10\x02\x32\xbc\x05\n\x0e\x42illingService\x12r\n\x11GetNodesAndPodsUI\x12).chalk.server.v1.GetNodesAndPodsUIRequest\x1a*.chalk.server.v1.GetNodesAndPodsUIResponse"\x06\x90\x02\x01\x80}\x0b\x12l\n\x0fGetNodesAndPods\x12\'.chalk.server.v1.GetNodesAndPodsRequest\x1a(.chalk.server.v1.GetNodesAndPodsResponse"\x06\x90\x02\x01\x80}\x0b\x12\x66\n\rGetUsageChart\x12%.chalk.server.v1.GetUsageChartRequest\x1a&.chalk.server.v1.GetUsageChartResponse"\x06\x90\x02\x01\x80}\x1c\x12x\n\x13GetUtilizationRates\x12+.chalk.server.v1.GetUtilizationRatesRequest\x1a,.chalk.server.v1.GetUtilizationRatesResponse"\x06\x90\x02\x01\x80}\x01\x12x\n\x13GetPodRequestCharts\x12+.chalk.server.v1.GetPodRequestChartsRequest\x1a,.chalk.server.v1.GetPodRequestChartsResponse"\x06\x90\x02\x01\x80}\x06\x12l\n\x0fSyncUtilization\x12\'.chalk.server.v1.SyncUtilizationRequest\x1a(.chalk.server.v1.SyncUtilizationResponse"\x06\x90\x02\x01\x80}\x1b\x42\x95\x01\n\x13\x63om.chalk.server.v1B\x0c\x42illingProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.billing_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\014BillingProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_BILLINGSERVICE"].methods_by_name["GetNodesAndPodsUI"]._options = None
    _globals["_BILLINGSERVICE"].methods_by_name["GetNodesAndPodsUI"]._serialized_options = b"\220\002\001\200}\013"
    _globals["_BILLINGSERVICE"].methods_by_name["GetNodesAndPods"]._options = None
    _globals["_BILLINGSERVICE"].methods_by_name["GetNodesAndPods"]._serialized_options = b"\220\002\001\200}\013"
    _globals["_BILLINGSERVICE"].methods_by_name["GetUsageChart"]._options = None
    _globals["_BILLINGSERVICE"].methods_by_name["GetUsageChart"]._serialized_options = b"\220\002\001\200}\034"
    _globals["_BILLINGSERVICE"].methods_by_name["GetUtilizationRates"]._options = None
    _globals["_BILLINGSERVICE"].methods_by_name["GetUtilizationRates"]._serialized_options = b"\220\002\001\200}\001"
    _globals["_BILLINGSERVICE"].methods_by_name["GetPodRequestCharts"]._options = None
    _globals["_BILLINGSERVICE"].methods_by_name["GetPodRequestCharts"]._serialized_options = b"\220\002\001\200}\006"
    _globals["_BILLINGSERVICE"].methods_by_name["SyncUtilization"]._options = None
    _globals["_BILLINGSERVICE"].methods_by_name["SyncUtilization"]._serialized_options = b"\220\002\001\200}\033"
    _globals["_USAGECHARTPERIOD"]._serialized_start = 1531
    _globals["_USAGECHARTPERIOD"]._serialized_end = 1647
    _globals["_USAGECHARTGROUPING"]._serialized_start = 1650
    _globals["_USAGECHARTGROUPING"]._serialized_end = 1782
    _globals["_GETUSAGECHARTREQUEST"]._serialized_start = 309
    _globals["_GETUSAGECHARTREQUEST"]._serialized_end = 573
    _globals["_GETUSAGECHARTRESPONSE"]._serialized_start = 575
    _globals["_GETUSAGECHARTRESPONSE"]._serialized_end = 644
    _globals["_GETUTILIZATIONRATESREQUEST"]._serialized_start = 646
    _globals["_GETUTILIZATIONRATESREQUEST"]._serialized_end = 674
    _globals["_GETUTILIZATIONRATESRESPONSE"]._serialized_start = 676
    _globals["_GETUTILIZATIONRATESRESPONSE"]._serialized_end = 756
    _globals["_GETNODESANDPODSREQUEST"]._serialized_start = 759
    _globals["_GETNODESANDPODSREQUEST"]._serialized_end = 969
    _globals["_GETNODESANDPODSRESPONSE"]._serialized_start = 972
    _globals["_GETNODESANDPODSRESPONSE"]._serialized_end = 1108
    _globals["_GETNODESANDPODSUIREQUEST"]._serialized_start = 1111
    _globals["_GETNODESANDPODSUIREQUEST"]._serialized_end = 1323
    _globals["_GETNODESANDPODSUIRESPONSE"]._serialized_start = 1326
    _globals["_GETNODESANDPODSUIRESPONSE"]._serialized_end = 1476
    _globals["_SYNCUTILIZATIONREQUEST"]._serialized_start = 1478
    _globals["_SYNCUTILIZATIONREQUEST"]._serialized_end = 1502
    _globals["_SYNCUTILIZATIONRESPONSE"]._serialized_start = 1504
    _globals["_SYNCUTILIZATIONRESPONSE"]._serialized_end = 1529
    _globals["_BILLINGSERVICE"]._serialized_start = 1785
    _globals["_BILLINGSERVICE"]._serialized_end = 2485
# @@protoc_insertion_point(module_scope)



from chalk._gen.chalk.server.v1 import environment_pb2 as _environment_pb2
from chalk._gen.chalk.server.v1 import team_pb2 as _team_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class BootstrapExtraSettingsEnvironment(_message.Message):
    __slots__ = ("settings",)
    class SettingsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...

    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    settings: _containers.ScalarMap[str, bool]
    def __init__(self, settings: _Optional[_Mapping[str, bool]] = ...) -> None: ...

class BootstrapExtraSettings(_message.Message):
    __slots__ = ("environments",)
    class GlobalEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bool
        def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...

    class EnvironmentsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: BootstrapExtraSettingsEnvironment
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[_Union[BootstrapExtraSettingsEnvironment, _Mapping]] = ...
        ) -> None: ...

    GLOBAL_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENTS_FIELD_NUMBER: _ClassVar[int]
    environments: _containers.MessageMap[str, BootstrapExtraSettingsEnvironment]
    def __init__(
        self, environments: _Optional[_Mapping[str, BootstrapExtraSettingsEnvironment]] = ..., **kwargs
    ) -> None: ...

class ParsedBootstrapConfigs(_message.Message):
    __slots__ = ("teams", "projects", "environments", "team_invites", "extra_settings", "global_pinned_base_image")
    TEAMS_FIELD_NUMBER: _ClassVar[int]
    PROJECTS_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENTS_FIELD_NUMBER: _ClassVar[int]
    TEAM_INVITES_FIELD_NUMBER: _ClassVar[int]
    EXTRA_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    GLOBAL_PINNED_BASE_IMAGE_FIELD_NUMBER: _ClassVar[int]
    teams: _containers.RepeatedCompositeFieldContainer[_team_pb2.Team]
    projects: _containers.RepeatedCompositeFieldContainer[_team_pb2.Project]
    environments: _containers.RepeatedCompositeFieldContainer[_environment_pb2.Environment]
    team_invites: _containers.RepeatedCompositeFieldContainer[_team_pb2.TeamInvite]
    extra_settings: BootstrapExtraSettings
    global_pinned_base_image: str
    def __init__(
        self,
        teams: _Optional[_Iterable[_Union[_team_pb2.Team, _Mapping]]] = ...,
        projects: _Optional[_Iterable[_Union[_team_pb2.Project, _Mapping]]] = ...,
        environments: _Optional[_Iterable[_Union[_environment_pb2.Environment, _Mapping]]] = ...,
        team_invites: _Optional[_Iterable[_Union[_team_pb2.TeamInvite, _Mapping]]] = ...,
        extra_settings: _Optional[_Union[BootstrapExtraSettings, _Mapping]] = ...,
        global_pinned_base_image: _Optional[str] = ...,
    ) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from chalk._gen.chalk.server.v1 import billing_pb2 as chalk_dot_server_dot_v1_dot_billing__pb2
from chalk._gen.chalk.server.v1 import pod_request_pb2 as chalk_dot_server_dot_v1_dot_pod__request__pb2


class BillingServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetNodesAndPodsUI = channel.unary_unary(
            "/chalk.server.v1.BillingService/GetNodesAndPodsUI",
            request_serializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsUIRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsUIResponse.FromString,
        )
        self.GetNodesAndPods = channel.unary_unary(
            "/chalk.server.v1.BillingService/GetNodesAndPods",
            request_serializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsResponse.FromString,
        )
        self.GetUsageChart = channel.unary_unary(
            "/chalk.server.v1.BillingService/GetUsageChart",
            request_serializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetUsageChartRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetUsageChartResponse.FromString,
        )
        self.GetUtilizationRates = channel.unary_unary(
            "/chalk.server.v1.BillingService/GetUtilizationRates",
            request_serializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetUtilizationRatesRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetUtilizationRatesResponse.FromString,
        )
        self.GetPodRequestCharts = channel.unary_unary(
            "/chalk.server.v1.BillingService/GetPodRequestCharts",
            request_serializer=chalk_dot_server_dot_v1_dot_pod__request__pb2.GetPodRequestChartsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_pod__request__pb2.GetPodRequestChartsResponse.FromString,
        )
        self.SyncUtilization = channel.unary_unary(
            "/chalk.server.v1.BillingService/SyncUtilization",
            request_serializer=chalk_dot_server_dot_v1_dot_billing__pb2.SyncUtilizationRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_billing__pb2.SyncUtilizationResponse.FromString,
        )


class BillingServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetNodesAndPodsUI(self, request, context):
        """GetNodesAndPodsUI returns the nodes and pods for the team by default,
        not just a single environment. To limit the scope, add filters to
        the request object.
        Use this endpoint going forward; GetNodesAndPods should be deprecated because
        it reuses PubSub types that are dangerous to update and are not intended for UI use.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetNodesAndPods(self, request, context):
        """GetNodesAndPods returns the nodes and pods for the team by default,
        not just a single environment. To limit the scope, add filters to
        the request object.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetUsageChart(self, request, context):
        """GetUsageChart shows the Chalk credit usage between a provided start and
        end period. The usage can be grouped by UsageChartPeriod for daily or
        monthly usage, and by UsageChartGrouping for instance type or cluster usage.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetUtilizationRates(self, request, context):
        """GetUtilizationRates returns the current utilization rates for all
        instance types.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetPodRequestCharts(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SyncUtilization(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_BillingServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "GetNodesAndPodsUI": grpc.unary_unary_rpc_method_handler(
            servicer.GetNodesAndPodsUI,
            request_deserializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsUIRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsUIResponse.SerializeToString,
        ),
        "GetNodesAndPods": grpc.unary_unary_rpc_method_handler(
            servicer.GetNodesAndPods,
            request_deserializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsResponse.SerializeToString,
        ),
        "GetUsageChart": grpc.unary_unary_rpc_method_handler(
            servicer.GetUsageChart,
            request_deserializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetUsageChartRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetUsageChartResponse.SerializeToString,
        ),
        "GetUtilizationRates": grpc.unary_unary_rpc_method_handler(
            servicer.GetUtilizationRates,
            request_deserializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetUtilizationRatesRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_billing__pb2.GetUtilizationRatesResponse.SerializeToString,
        ),
        "GetPodRequestCharts": grpc.unary_unary_rpc_method_handler(
            servicer.GetPodRequestCharts,
            request_deserializer=chalk_dot_server_dot_v1_dot_pod__request__pb2.GetPodRequestChartsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_pod__request__pb2.GetPodRequestChartsResponse.SerializeToString,
        ),
        "SyncUtilization": grpc.unary_unary_rpc_method_handler(
            servicer.SyncUtilization,
            request_deserializer=chalk_dot_server_dot_v1_dot_billing__pb2.SyncUtilizationRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_billing__pb2.SyncUtilizationResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("chalk.server.v1.BillingService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class BillingService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetNodesAndPodsUI(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.BillingService/GetNodesAndPodsUI",
            chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsUIRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsUIResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetNodesAndPods(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.BillingService/GetNodesAndPods",
            chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_billing__pb2.GetNodesAndPodsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetUsageChart(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.BillingService/GetUsageChart",
            chalk_dot_server_dot_v1_dot_billing__pb2.GetUsageChartRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_billing__pb2.GetUsageChartResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetUtilizationRates(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.BillingService/GetUtilizationRates",
            chalk_dot_server_dot_v1_dot_billing__pb2.GetUtilizationRatesRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_billing__pb2.GetUtilizationRatesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetPodRequestCharts(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.BillingService/GetPodRequestCharts",
            chalk_dot_server_dot_v1_dot_pod__request__pb2.GetPodRequestChartsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_pod__request__pb2.GetPodRequestChartsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SyncUtilization(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.BillingService/SyncUtilization",
            chalk_dot_server_dot_v1_dot_billing__pb2.SyncUtilizationRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_billing__pb2.SyncUtilizationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from os import PathLike
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
)

import pyarrow as pa

if TYPE_CHECKING:
    import sqlalchemy.ext.asyncio
    from sqlalchemy.engine import Connection

from chalk.features import Feature
from chalk.sql._internal.chalk_query import ChalkQuery
from chalk.sql._internal.query_execution_parameters import QueryExecutionParameters
from chalk.sql._internal.sql_source import _ENABLE_ADD_TO_SQL_SOURCE_REGISTRIES  # pyright: ignore[reportPrivateUsage]
from chalk.sql._internal.sql_source import BaseSQLSource
from chalk.sql._internal.string_chalk_query import StringChalkQuery
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.sql.protocols import BaseSQLSourceProtocol, ChalkQueryProtocol, StringChalkQueryProtocol
from chalk.utils.datasource_tags import tag_matches
from chalk.utils.missing_dependency import missing_dependency_exception


class SQLSourceGroup(BaseSQLSourceProtocol):
    registry: ClassVar[List["SQLSourceGroup"]] = []

    def __init__(self, name: str, default: BaseSQLSource, tagged_sources: Mapping[str, BaseSQLSource]):
        super().__init__()

        if hasattr(default, "kind"):
            self.kind = default.kind

        self.name = name
        self._default = default
        self._tagged_sources = tagged_sources
        if _ENABLE_ADD_TO_SQL_SOURCE_REGISTRIES.get():
            self.registry.append(self)

        try:
            import sqlalchemy
        except ImportError:
            raise missing_dependency_exception("chalkpy[sql]")
        del sqlalchemy  # unused

    def to_json(self) -> Dict[str, str]:
        return self._default.to_json()

    def get_sqlglot_dialect(self) -> Optional[str]:
        return self._default.get_sqlglot_dialect()

    def query_string(
        self,
        query: str,
        fields: Optional[Mapping[str, Union[Feature, str, Any]]] = None,
        args: Optional[Mapping[str, object]] = None,
    ) -> StringChalkQueryProtocol:

        default_response = self._default.query_string(query, fields, args)

        if isinstance(default_response, StringChalkQuery):
            default_response._source = self  # pyright: ignore[reportPrivateUsage]

        return default_response

    def query_sql_file(
        self,
        path: Union[str, bytes, PathLike],
        fields: Optional[Mapping[str, Union[Feature, str, Any]]] = None,
        args: Optional[Mapping[str, object]] = None,
    ) -> StringChalkQueryProtocol:
        default_response = self._default.query_sql_file(path, fields, args)

        if isinstance(default_response, StringChalkQuery):
            default_response._source = self  # pyright: ignore[reportPrivateUsage]

        return default_response

    def query(self, *entities: Any) -> ChalkQueryProtocol:
        default_response = self._default.query(*entities)

        if isinstance(default_response, ChalkQuery):
            default_response._source = self  # pyright: ignore[reportPrivateUsage]

        return default_response

    def execute_query(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: "Optional[Connection]" = None,
        query_execution_parameters: Optional[QueryExecutionParameters] = None,
    ) -> Iterable[pa.RecordBatch]:
        source = self.select_matching_source(query_execution_parameters=query_execution_parameters)
        return source.execute_query(
            finalized_query=finalized_query,
            columns_to_features=columns_to_features,
            connection=connection,
            query_execution_parameters=query_execution_parameters,
        )

    async def async_execute_query(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: "Optional[sqlalchemy.ext.asyncio.AsyncConnection]" = None,
        query_execution_parameters: Optional[QueryExecutionParameters] = None,
    ) -> AsyncIterable[pa.RecordBatch]:
        source = self.select_matching_source(query_execution_parameters=query_execution_parameters)
        return source.async_execute_query(
            finalized_query=finalized_query,
            columns_to_features=columns_to_features,
            connection=connection,
            query_execution_parameters=query_execution_parameters,
        )

    def get_engine(self) -> "sqlalchemy.engine.Engine":
        raise NotImplementedError("SQLSourceGroup does not support get_engine")

    def select_matching_source(self, query_execution_parameters: Optional[QueryExecutionParameters]) -> BaseSQLSource:
        if query_execution_parameters is None:
            return self._default

        candidate = None

        for tag, source in self._tagged_sources.items():
            if tag_matches(datasource_tag=tag, context_tags=query_execution_parameters.tags):
                if candidate is not None:
                    raise ValueError(
                        f"Ambiguous tags in datasource selection; multiple sources match: {source} and {candidate}. Provided tags were: {query_execution_parameters.tags}."
                    )
                candidate = source

        return candidate or self._default



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.aggregate.v1.service_pb2 import (
    GetAggregatesRequest,
    GetAggregatesResponse,
    PlanAggregateBackfillRequest,
    PlanAggregateBackfillResponse,
)
from chalk._gen.chalk.common.v1.online_query_pb2 import (
    OnlineQueryBulkRequest,
    OnlineQueryBulkResponse,
    OnlineQueryMultiRequest,
    OnlineQueryMultiResponse,
    OnlineQueryRequest,
    OnlineQueryResponse,
    UploadFeaturesBulkRequest,
    UploadFeaturesBulkResponse,
)
from chalk._gen.chalk.common.v1.upload_features_pb2 import (
    UploadFeaturesRequest,
    UploadFeaturesResponse,
)
from chalk._gen.chalk.engine.v1.query_server_pb2 import (
    PingRequest,
    PingResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class QueryServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    Ping: UnaryUnaryMultiCallable[
        PingRequest,
        PingResponse,
    ]
    OnlineQuery: UnaryUnaryMultiCallable[
        OnlineQueryRequest,
        OnlineQueryResponse,
    ]
    OnlineQueryBulk: UnaryUnaryMultiCallable[
        OnlineQueryBulkRequest,
        OnlineQueryBulkResponse,
    ]
    OnlineQueryMulti: UnaryUnaryMultiCallable[
        OnlineQueryMultiRequest,
        OnlineQueryMultiResponse,
    ]
    UploadFeaturesBulk: UnaryUnaryMultiCallable[
        UploadFeaturesBulkRequest,
        UploadFeaturesBulkResponse,
    ]
    UploadFeatures: UnaryUnaryMultiCallable[
        UploadFeaturesRequest,
        UploadFeaturesResponse,
    ]
    PlanAggregateBackfill: UnaryUnaryMultiCallable[
        PlanAggregateBackfillRequest,
        PlanAggregateBackfillResponse,
    ]
    """PlanAggregateBackfill determines the estimated resources needed to backfill
    an aggregate.

    This method is a duplicate of the PlanAggregateBackfill method
    in the query_server.proto file. We should remove the query_server.proto method
    and move that request to this service instead.
    buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
    """
    GetAggregates: UnaryUnaryMultiCallable[
        GetAggregatesRequest,
        GetAggregatesResponse,
    ]
    """This method is a duplicate of the PlanAggregateBackfill method
    in the query_server.proto file. We should remove the query_server.proto method
    and move that request to this service instead.
    buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
    """

class QueryServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def Ping(
        self,
        request: PingRequest,
        context: ServicerContext,
    ) -> PingResponse: ...
    @abstractmethod
    def OnlineQuery(
        self,
        request: OnlineQueryRequest,
        context: ServicerContext,
    ) -> OnlineQueryResponse: ...
    @abstractmethod
    def OnlineQueryBulk(
        self,
        request: OnlineQueryBulkRequest,
        context: ServicerContext,
    ) -> OnlineQueryBulkResponse: ...
    @abstractmethod
    def OnlineQueryMulti(
        self,
        request: OnlineQueryMultiRequest,
        context: ServicerContext,
    ) -> OnlineQueryMultiResponse: ...
    @abstractmethod
    def UploadFeaturesBulk(
        self,
        request: UploadFeaturesBulkRequest,
        context: ServicerContext,
    ) -> UploadFeaturesBulkResponse: ...
    @abstractmethod
    def UploadFeatures(
        self,
        request: UploadFeaturesRequest,
        context: ServicerContext,
    ) -> UploadFeaturesResponse: ...
    @abstractmethod
    def PlanAggregateBackfill(
        self,
        request: PlanAggregateBackfillRequest,
        context: ServicerContext,
    ) -> PlanAggregateBackfillResponse:
        """PlanAggregateBackfill determines the estimated resources needed to backfill
        an aggregate.

        This method is a duplicate of the PlanAggregateBackfill method
        in the query_server.proto file. We should remove the query_server.proto method
        and move that request to this service instead.
        buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
        """
    @abstractmethod
    def GetAggregates(
        self,
        request: GetAggregatesRequest,
        context: ServicerContext,
    ) -> GetAggregatesResponse:
        """This method is a duplicate of the PlanAggregateBackfill method
        in the query_server.proto file. We should remove the query_server.proto method
        and move that request to this service instead.
        buf:lint:ignore RPC_REQUEST_RESPONSE_UNIQUE
        """

def add_QueryServiceServicer_to_server(servicer: QueryServiceServicer, server: Server) -> None: ...



from __future__ import annotations

import collections
from typing import TYPE_CHECKING, Dict, List

from chalk.parsed.duplicate_input_gql import GraphLogSeverity, MetadataSettings, UpsertFeatureGQL, UpsertResolverGQL
from chalk.utils.string import to_camel_case

if TYPE_CHECKING:
    from chalk.parsed._graph_validation import ClientLogBuilder


MissingMetadata = Dict[GraphLogSeverity, List[str]]

METADATA_STRING_TO_FEATURE_ATTRIBUTE_MAP: Dict[str, str] = {
    "owner": "owner",
    "tags": "tags",
    "description": "description",
}

METADATA_STRING_TO_RESOLVER_ATTRIBUTE_MAP: Dict[str, str] = {
    "owner": "owner",
    "tags": "tags",
    "description": "doc",
    "doc": "doc",
    "docstring": "doc",
}


class MissingMetadataFeature:
    def __init__(self, feature: UpsertFeatureGQL, missing_metadata: MissingMetadata):
        super().__init__()
        self.feature = feature
        self.missing_metadata = missing_metadata


class MissingMetadataResolver:
    def __init__(self, resolver: UpsertResolverGQL, missing_metadata: MissingMetadata):
        super().__init__()
        self.resolver = resolver
        self.missing_metadata = missing_metadata


def build_feature_log_for_each_severity(
    builder: "ClientLogBuilder", missing_metadata_features: list[MissingMetadataFeature]
):
    if not missing_metadata_features:
        return

    for severity in GraphLogSeverity:
        log_or_none = get_missing_metadata_features_log(
            missing_metadata_features=missing_metadata_features, severity=GraphLogSeverity(severity)
        )
        if log_or_none is None:
            continue
        header, subheader = log_or_none
        builder.add_log(header=header, subheader=subheader, severity=GraphLogSeverity(severity))


def build_resolver_log_for_each_severity(
    builder: "ClientLogBuilder", missing_metadata_resolvers: list[MissingMetadataResolver]
):
    if not missing_metadata_resolvers:
        return

    for severity in GraphLogSeverity:
        for resolver in missing_metadata_resolvers:
            log_or_none = get_missing_metadata_resolver_log(
                missing_metadata_resolver=resolver, severity=GraphLogSeverity(severity)
            )
            if log_or_none is None:
                continue
            header, subheader = log_or_none
            builder.add_log(header=header, subheader=subheader, severity=GraphLogSeverity(severity))


def get_missing_metadata(
    entity: UpsertFeatureGQL | UpsertResolverGQL,
    settings: list[MetadataSettings],
    attribute_map: dict[str, str],
) -> MissingMetadata:
    res: MissingMetadata = collections.defaultdict(list)
    for setting in settings:
        metadata_name = setting.name
        try:
            severity = GraphLogSeverity(setting.missing.upper())
        except:
            """If this fails, _validate_metadata_config should log the error"""
            continue
        if metadata_name in attribute_map:
            attribute_name = attribute_map[metadata_name]
            value = getattr(entity, attribute_name)
            is_missing = value is None or (isinstance(value, list) and len(value) == 0)
            if is_missing:
                res[severity].append(metadata_name)
    return res


def validate_feature_metadata(
    settings: list[MetadataSettings],
    namespace_features: list[UpsertFeatureGQL],
) -> list[MissingMetadataFeature]:
    wrapped_features = []
    for nf in namespace_features:
        missing_metadata = get_missing_metadata(
            entity=nf,
            attribute_map=METADATA_STRING_TO_FEATURE_ATTRIBUTE_MAP,
            settings=settings,
        )
        if not missing_metadata:
            continue
        wrapped_features.append(MissingMetadataFeature(feature=nf, missing_metadata=missing_metadata))

    return wrapped_features


def get_missing_metadata_features_log(
    missing_metadata_features: list[MissingMetadataFeature], severity: GraphLogSeverity
) -> tuple[str, str] | None:
    if not missing_metadata_features:
        return None

    missing_messages = []
    feature_str = "feature"
    missing_metadata_header_str = "missing metadata"
    max_name_len = max([len(w.feature.id.name) for w in missing_metadata_features])
    feature_column_width = max(max_name_len, len(feature_str)) + 1
    get_padding = lambda s: feature_column_width - len(s)

    first_feature = missing_metadata_features[0].feature
    header = f'"{to_camel_case(first_feature.id.namespace)}" features have missing metadata'
    subheader = f"  Filepath: {first_feature.namespacePath}\n\n"

    for wrapper in missing_metadata_features:
        missing_metadata = wrapper.missing_metadata.get(severity)
        if not missing_metadata:
            continue
        padding_1 = get_padding(wrapper.feature.id.name)
        missing_metadata_str = ", ".join(missing_metadata)
        missing_messages.append(f"      {wrapper.feature.id.name}{' ' * padding_1}: {missing_metadata_str}")

    if not missing_messages:
        return None

    padding_2 = get_padding(feature_str)
    subheader += f"      {'-' * len(feature_str)}{' ' * padding_2}  {'-' * len(missing_metadata_header_str)}\n"
    subheader += f"      {feature_str}{' ' * padding_2}  {missing_metadata_header_str}\n"
    subheader += f"      {'-' * len(feature_str)}{' ' * padding_2}  {'-' * len(missing_metadata_header_str)}\n"
    subheader += "\n".join(missing_messages)

    return header, subheader


def get_missing_metadata_resolver_log(
    missing_metadata_resolver: MissingMetadataResolver, severity: GraphLogSeverity
) -> tuple[str, str] | None:
    missing_metadata = missing_metadata_resolver.missing_metadata.get(severity)
    if missing_metadata is None:
        return
    missing_metadata_str = ", ".join(missing_metadata)
    resolver_str = "resolver"
    missing_metadata_header_str = "missing metadata"
    resolver = missing_metadata_resolver.resolver
    max_name_len = len(resolver.fqn)
    feature_column_width = max(max_name_len, len(resolver_str)) + 1
    get_padding = lambda s: feature_column_width - len(s)
    padding_1 = get_padding(resolver.fqn)

    header = f'Resolver "{resolver.fqn}" has missing metadata'
    subheader = f"  Filepath: {missing_metadata_resolver.resolver.filename}\n\n"
    message = f"      {resolver.fqn}{' ' * padding_1}: {missing_metadata_str}"

    padding_2 = get_padding(resolver_str)
    subheader += f"      {'-' * len(resolver_str)}{' ' * padding_2}  {'-' * len(missing_metadata_header_str)}\n"
    subheader += f"      {resolver_str}{' ' * padding_2}  {missing_metadata_header_str}\n"
    subheader += f"      {'-' * len(resolver_str)}{' ' * padding_2}  {'-' * len(missing_metadata_header_str)}\n"
    subheader += message

    return header, subheader
